/*
GLDM - Graphical Liberty Display Manager

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software Foundation,
Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include <X11/Xlib.h>
#include <sys/wait.h>

#include <err.h>
#include <signal.h>
#include <unistd.h>

#include "interfaces/simplesdlinterface.h"
#include "interfaces/lunaticinterface.h"
#include "dummyauthprovider.h"
#include "unixauthprovider.h"
#include "utils.h"


#define INTERFACE LunaticInterface
//#define INTERFACE SimpleSDLInterface

/// \bug FIXME ASAP this must be randomized
static std::string mcookie = "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa";
static std::string authfile = "/tmp/.gldm-auth-file";
#if (XEPHYR)
static std::string displayname = ":25";
#else
static std::string displayname = ":1";
#endif

Display *g_disp=NULL;

static void registerAPs()
{
    AuthProvider*ap;

    //ap = new DummyAuthProvider(); // stranger danger! use only for debugging
    //AuthProvider::registerProvider(ap);

    ap = new UnixAuthProvider();
    AuthProvider::registerProvider(ap);

}

static void createServerAuth()
{
    remove(authfile.c_str());
    putenv(StrConcat("XAUTHORITY=", authfile.c_str()));
    putenv(StrConcat("DISPLAY=", displayname.c_str()));
    // FIXME BUG don't use system
    // see http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=529306
    std::string cmd = std::string("/usr/bin/xauth") + " -q -f " + authfile + " add " + displayname + " . " + mcookie;
    system(cmd.c_str());

}

static pid_t startX11()
{
    pid_t x11pid;

    printf("Launching X server\n");
    if ((x11pid=fork())==0)
    {
        // forked process

        #if (XEPHYR)
        const char   server[] = "/usr/bin/Xephyr";
        const char*  argv[32];
        putenv("DISPLAY=:0");
        int argc=0;
        printf("(Remark: If Xephyr won't run, remember you have to set xhost +.)\n");
        #else
        const char   server[] = "/usr/bin/X";
        const char*  argv[32];
        int argc=0;
        #endif

        argv[argc++]=strrchr(server, '/')+1;
        printf("argv[0]: %s\n", argv[0]);
        #if (!XEPHYR)
        argv[argc++]="vt8";
        #endif
        argv[argc++]="-br";
        argv[argc++]="-ac"; // FIXME stranger danger! this allows anyone, without auth, to connect to our server.
        //argv[argc++]="-auth";
        //argv[argc++]=authfile.c_str();
        argv[argc++]=displayname.c_str();
        #if (XEPHYR)
        argv[argc++]="-screen";
        argv[argc++]="800x600";
        #endif
        argv[argc++]=0;
        execv(server, (char**)argv);
        return 0; // never happens
    }
    else
    {
        return x11pid;
    }
}


void sdl_to_root_window_hack(Display* disp)
{
    return;
    // hack for sdl  using root window
    //http://lists.libsdl.org/pipermail/sdl-libsdl.org/1999-October/003391.html

    if(disp) {
        char SDL_windowhack[32];
        sprintf(SDL_windowhack,"SDL_WINDOWID=0x%x", (unsigned int)DefaultRootWindow(disp));

        putenv(SDL_windowhack);
        printf("SDL_windowhack: %s\n", SDL_windowhack);
        // alternative: export SDL_WINDOWID=`xwininfo -root|grep "id:"|sed 's/^.*id: //'|sed 's/ (.*$//'`
        // this is from stackoverflow
    }

}

int debugMain()
{
    BaseInterface*ui;

    registerAPs();

    g_disp = XOpenDisplay(""); //displayname.c_str());
    sdl_to_root_window_hack(g_disp);
//    SDL_Init(SDL_INIT_VIDEO);



    for (;;) {
        ui = new INTERFACE();
        ui->showSelf();
        while(!ui->isFinished() && !ui->isCanceled())
        {
            static uint32_t lastUpdate = SDL_GetTicks();
            uint32_t now = SDL_GetTicks();
            float delta = (SDL_GetTicks()-lastUpdate)/1000.;

            ui->paintSelf();
            if(delta>0)
                ui->updateSelf(delta);
            lastUpdate = now;
        }
        /*
        for(ui->showSelf();!ui->isFinished();ui->updateSelf(0.1))
        {
            ui->paintSelf();
        }*/
        ui->hideSelf();

        if (!ui->isCanceled())
        {

            AuthProvider* ap = ui->getSucceedingAuthProvider();

            if (!ap->switchUser(ui->getLogin(), ui->getPassword(), getenv("DISPLAY"), "gnome-session")) //"xterm -e \"whoami;env;read\""))
            {
                printf("Fail in switching user\n");
            }
        }

        bool wasCanceled = ui->isCanceled();

        delete ui;
        if (wasCanceled)
        {
            break;
        }
    }

    return 0;
}

int releaseMain()
{
    BaseInterface*ui;

    registerAPs();
    createServerAuth();
    for (;;) {

        pid_t pid = startX11();
        if (pid==0)  // forked child finished... although it shouldnt happen
            return 0;
        printf("------ STARTED X... waiting -----------\n");
        sleep(2);



        // This hack keeps an X connection open during lifetime of the session.
        // This prevents crash of buggy X11.
        // See Debian bug reports, primarily #549085, with same crash as in
        // #548716.
        //
        // http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=549085
        // http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=548716

        if((g_disp = XOpenDisplay(displayname.c_str())) == 0)
        {
            kill(pid, SIGTERM);
            wait(NULL); // just collect return info
            errx(-1, "Could not open display");
        }
        // End hack.


        sdl_to_root_window_hack(g_disp);

        printf("Back in main, launching interface\n");
        ui = new INTERFACE();
        ui->showSelf();

        float runningtime=0;
        int oldticks=SDL_GetTicks();
        while(!ui->isFinished() && !ui->isCanceled())
        {
            int newticks;

            ui->paintSelf();
            newticks=SDL_GetTicks();
            if(newticks!=oldticks)
            {
                runningtime+=(newticks-oldticks)/1000.;
                ui->updateSelf((newticks-oldticks)/1000.);
            }
            oldticks=newticks;
        }
        ui->hideSelf();


        if (!ui->isCanceled())
        {
            AuthProvider* ap = ui->getSucceedingAuthProvider();
            printf("SWITCHING USER to %s\n",ui->getLogin().c_str());
            if (!ap->switchUser(ui->getLogin(), ui->getPassword(), getenv("DISPLAY"), "gnome-session")) //"xterm"))//" -e \"whoami;env;read\""))
            {
                warnx("Fail in switching user");
            }
        }
        bool wasCanceled = ui->isCanceled();

        delete ui;
        XCloseDisplay(g_disp); // Hack as described above near XOpenDisplay();
        kill(pid, SIGTERM);
        wait(NULL); // just collect return info

        if (wasCanceled)
        {
            printf("Canceled\n");
            break;
        }

        // let's wait a while for x to shut down
        sleep(2);
    }

    return 0;

}

int main ()
{
    #ifndef THEMEDEBUG
    return releaseMain();
    #endif
    return debugMain();
}
